# Copyright 2010, 2011 Martin Pool
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

import os
import webbrowser

import glib
import gtk.gdk
import gtk.glade

from twisted.internet import reactor
from twisted.internet.defer import (
    Deferred,
    )
from twisted.python import (
    log,
    )

import blurtr
from blurtr.feed import (
    open_feed_from_location,
    )
from blurtr.cache import (
    ImageCache,
    )
from blurtr.net import (
    FeedImageToLoader,
    TumblrFeedProtocol,
    request_url_to_protocol,
    )


def load_glade_xml(filename):
    """Return the Glade XML object holding window definitions."""
    path = os.path.join(os.path.dirname(blurtr.__file__), filename)
    return gtk.glade.XML(path)


glade_xml = load_glade_xml('ui.glade')


class FeedCursor(object):
    """Holds state for navigating through the feed"""

    def __init__(self, feed, window):
        self.feed = feed 
        self.window = window
        self.photo_posts = [
            p for p in self.feed.posts if p.type == 'photo']
        self.index = None

    @property
    def current_post(self):
        return self.photo_posts[self.index]

    @property
    def num_photo_posts(self):
        return len(self.photo_posts)

    def goto_first(self):
        self.goto_index(0)

    def goto_next(self):
        self.goto_index(self.index + 1)

    def goto_previous(self):
        self.goto_index(self.index - 1)

    def goto_last(self):
        self.goto_index(self.num_photo_posts - 1)

    def goto_index(self, index):
        if index >= 0 and index < self.num_photo_posts and index != self.index:
            self.index = index
            p = self.photo_posts[self.index]
            print 'on post %d/%d' % (self.index + 1, self.num_photo_posts)
            self.window.show_photo_post(p)
            self.window.set_position_indicator(self.index, self.num_photo_posts)


def affine_scale_pixbuf(pixbuf, bound_width, bound_height, allow_grow):
    pw = pixbuf.get_width()
    ph = pixbuf.get_height()
    hr = ph / bound_height      # true division, pep0238
    wr = pw / bound_width
    scale_ratio = max(hr, wr)
    if scale_ratio > 1 or allow_grow:
        to_height = ph / scale_ratio
        to_width = pw / scale_ratio
        print 'scale image from %dx%d to %dx%d by ratio %s' % (
            pw, ph, to_width, to_height, scale_ratio)
        scaled_pixbuf = pixbuf.scale_simple(
            int(to_width), int(to_height), gtk.gdk.INTERP_BILINEAR)
        return scaled_pixbuf
    else:
        return pixbuf


class ImageView(object):
    """Manages the image view itself."""

    def __init__(self, image, cursor):
        self.image = image
        self.pixbuf = None
        self.last_scale_height = self.last_scale_width = 0
        self.cursor = cursor

    def on_main_image_expose_event(self, widget, event):
        print 'expose'
        self._rescale_image()

    def on_main_image_key_release_event(self, from_widget, event):
        print 'key released: %r %r from %r' % (event, event.keyval, from_widget)
        keyname = gtk.gdk.keyval_name(event.keyval)
        print 'name:', keyname
        if event.keyval == ord(' '):
            self.cursor.goto_next()
        elif keyname == 'Page_Down':
            self.cursor.goto_next()
        elif keyname == 'Page_Up':
            self.cursor.goto_previous()

    def _rescale_image(self):
        print "rescale image"
        if self.pixbuf is None:
            return
        alloc = self.image.allocation
        if (self.last_scale_height == alloc.height
            and self.last_scale_width == alloc.width):
            # can be a bit smarter and avoid rescaling when the change won't
            # affect the scale factor
            print "no actual change in size"
            return
        self.set_image_from_pixbuf(self.pixbuf)

    def set_image_from_pixbuf(self, pixbuf):
        self.pixbuf = pixbuf
        alloc = self.image.allocation
        # if way too small, just don't bother
        if alloc.width < 4 or alloc.height < 4:
            return
        scaled_pixbuf = affine_scale_pixbuf(
            pixbuf, alloc.width, alloc.height, False)
        self.image.set_from_pixbuf(scaled_pixbuf)
        self.image.set_size_request(0, 0)
        self.last_scale_width = alloc.width
        self.last_scale_height = alloc.height


class StatusbarView(object):

    def __init__(self, statusbar):
        self.statusbar = statusbar
        self.contexts = {}

    def set_message(self, context_name, message):
        context = self.statusbar.get_context_id(context_name)
        if message is None:
            self.statusbar.pop(context)
        else:
            self.statusbar.push(context, message)


class BlurtrWindow(object):
    """Object that views tumblr content etc.
    """

    def __init__(self):
        self.glade_xml = glade_xml
        self.get_widget = glade_xml.get_widget
        glade_xml.signal_autoconnect(self)
        for name in ['main_window', 'main_image',
            'tb_top', 'tb_bottom', 'tb_up', 'tb_down',
            'statusbar',
            'viewport',
            ]:
            setattr(self, name, self.get_widget(name))
        self.image_view = ImageView(self.main_image, None)
        glade_xml.signal_autoconnect(self.image_view)
        self.main_window.show_all()
        self.image_cache = ImageCache()
        grey = gtk.gdk.Color(0.2, 0.2, 0.2)
        self.viewport.modify_bg(gtk.STATE_NORMAL, grey)
        self.statusbar_view = StatusbarView(self.statusbar)

    def on_main_window_destroy(self, what):
        reactor.stop()

    def on_tb_top_clicked(self, what):
        self.cursor.goto_first()

    def on_tb_down_clicked(self, what):
        self.cursor.goto_next()

    def on_tb_open_post_clicked(self, what):
        webbrowser.open_new(self.cursor.current_post.url)

    def on_tb_up_clicked(self, what):
        self.cursor.goto_previous()

    def on_tb_bottom_clicked(self, what):
        self.cursor.goto_last()

    def on_tb_refresh_clicked(self, widget):
        self.show_tumblr_feed(self.feed_url)

    def set_position_indicator(self, index, num_posts):
        # GNOME HIG says including the application name is normally NOT
        # recommended.
        #
        # For a dashboard, we have an origin-feed title and name per post; for
        # a regular feed only one for the whole thing.
        post = self.cursor.current_post
        title = post.title or self.feed.title
        name = post.name or self.feed.name
        self.main_window.set_title(
            "%s [%s] (%s) (%d/%d)"
            % (title, name, post.date, index + 1, num_posts))
        can_up = (index > 0)
        can_down = (index + 1 < num_posts)
        self.tb_top.set_sensitive(can_up)
        self.tb_up.set_sensitive(can_up)
        self.tb_down.set_sensitive(can_down)
        self.tb_bottom.set_sensitive(can_down)

    def show_tumblr_feed(self, feed_url):
        self.feed_url = feed_url
        if feed_url.startswith('http://'):
            self.set_busy(True)
            self.statusbar_view.set_message('loading',
                'Loading feed %s' % feed_url)
            finished = Deferred()
            finished.addCallback(self._show_feed_object)
            d = request_url_to_protocol(feed_url,
                TumblrFeedProtocol(finished))
            d.addErrback(self._failed_to_load)
        else:
            self._show_feed_object(open_feed_from_location(feed_url))

    def _failed_to_load(self, failure):
        self.set_busy(False)
        self.statusbar_view.set_message('loading', None)
        self.statusbar_view.set_message('failed', failure.getErrorMessage().strip())

    def _show_feed_object(self, feed):
        print 'showing feed %r' % feed
        self.statusbar_view.set_message('loading', None)
        self.feed = feed
        self.image_view.cursor = self.cursor = FeedCursor(self.feed, self)
        self.statusbar_view.set_message('failed', None)
        self.cursor.goto_first()

    def show_photo_post(self, p):
        """Show the image from a single photo post."""
        for max_size, photo_url in p.photo_urls:
            # For now, just always use the maximum-sized image, which always
            # comes first.
            break
        print 'open photo %s (max size %d)' % (photo_url, max_size)
        self.show_image_url(photo_url)

    def show_image_url(self, url):
        self.set_busy(True)
        self.statusbar_view.set_message('loading',
            'Loading photo %s...' % url)
        cache_filename = self.image_cache.get_present_cache_filename(url)
        if cache_filename is not None:
            pixbuf = gtk.gdk.pixbuf_new_from_file(cache_filename)
            self.set_image_from_pixbuf(pixbuf)
            self._finished_loading(None)
            print 'loaded %s from cache' % (url,)
            return
        else:
            self._background_load_image(url)

    def _finished_loading(self, result):
        # not necessarily successfully
        print 'finished loading'
        self.set_busy(False)
        self.statusbar_view.set_message('loading', None)
        return result

    def _background_load_image(self, url):
        loader = gtk.gdk.PixbufLoader()
        after = Deferred()
        after.addCallback(self._set_image_from_loader)
        after.addErrback(log.err)
        after.addBoth(self._finished_loading)
        feeder = FeedImageToLoader(url, after, loader, self.image_cache)
        request_url_to_protocol(url, feeder)

    def _set_image_from_loader(self, pixbuf_loader):
        # called after loading completes
        print 'done loading image'
        pixbuf = pixbuf_loader.get_pixbuf()
        self.set_image_from_pixbuf(pixbuf)
        self.statusbar_view.set_message('failed', None)

    def set_icon_from_pixbuf(self, pixbuf):
        print "set window icon"
        # Although gtk_window_set_icon says you shouldn't pre-scale, it turns
        # out that it actually complains about very large icons.  However, we
        # just scale to one intermediate-sized image, and let gtk go from
        # there, to save CPU cycles.
        self.main_window.set_icon_list(
            *map(
                lambda x: affine_scale_pixbuf(pixbuf, x, x, False),
                [256]))

    def set_image_from_pixbuf(self, pixbuf):
        self.set_icon_from_pixbuf(pixbuf)
        self.image_view.set_image_from_pixbuf(pixbuf)

    def set_busy(self, is_busy):
        if is_busy:
            self.set_cursor(gtk.gdk.WATCH)
        else:
            self.set_cursor(None)

    def set_cursor(self, cursor_type):
        if cursor_type is None:
            new_cursor = None
        else:
            new_cursor = gtk.gdk.Cursor(cursor_type)
        self.main_window.window.set_cursor(new_cursor)
